{$I DFS.INC}  { Standard defines for all Delphi Free Stuff components }

{-----------------------------------------------------------------------------}
{ TGradientForm v1.56                                                         }
{ A form to provide gradient filled caption bars ala Microsoft Office.        }
{ You will notice that some of the initial comment characters are followed by }
{ a colon, and those sometimes contains some odd looking things that resemble }
{ HTML codes.  These comments are used by the Time2Help application that I    }
{ used to build the help file.                                                }
{-----------------------------------------------------------------------------}
{ Copyright 1998, Brad Stowers.  All Rights Reserved.                         }
{ This component can be freely used and distributed in commercial and private }
{ environments, provided this notice is not modified in any way and there is  }
{ no charge for it other than nomial handling fees.  Contact me directly for  }
{ modifications to this agreement.                                            }
{-----------------------------------------------------------------------------}
{ Feel free to contact me if you have any questions, comments or suggestions  }
{ at bstowers@pobox.com.                                                      }
{ The lateset version will always be available on the web at:                 }
{   http://www.pobox.com/~bstowers/delphi/                                    }
{ See GradForm.txt for notes, known issues, and revision history.             }
{-----------------------------------------------------------------------------}
{ Date last modified:  May 14, 1998                                           }
{-----------------------------------------------------------------------------}


{: This unit provides the TGradientForm class, and all supporting elements. }
unit GradForm;

{$IFNDEF DFS_WIN32}
  Error!  This unit is only available for Win32.
{$ENDIF}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

const
  {: The minimum number of colors that can be assigned to the
     <See Property=TGradientForm.GradientColors Text=GradientColors> property.
     Any less than 8 colors doesn't look much like a gradient.
     <Related A=MAX_GRADIENT_COLORS;DEF_GRADIENT_COLORS>
  }
  MIN_GRADIENT_COLORS      = 8;
  {: The maximum number of colors that can be assigned to the
     <See Property=TGradientForm.GradientColors Text=GradientColors> property.
     Any more than 512 colors is not noticeable, and just slows the painting
     down. <Related A=MIN_GRADIENT_COLORS;DEF_GRADIENT_COLORS> }
  MAX_GRADIENT_COLORS      = 512;
  {: The default number of colors for the
     <See Property=TGradientForm.GradientColors Text=GradientColors> property.
     This is a good compromise between speed and appearance.
     <Related A=MAX_GRADIENT_COLORS;MIN_GRADIENT_COLORS>}
  DEF_GRADIENT_COLORS      = 64;
  DEF_CAPTION_TEXT_COLOR   = clWhite;
  DEF_GRADIENT_START_COLOR = clBlack;
  DEF_GRADIENT_STOP_COLOR  = clActiveCaption;

type
  {: This enumerated type is used by the
     <See Property=TGradientForm.PaintGradient Text=PaintGradient> property to
     indicate when the caption should be painted as a gradient.<BR>
     <UL>
     <LI>gfpAlways <TAB> The gradient should always be drawn.
     <LI>gfpActive <TAB> Only draw the gradient when the form is active.
     <LI>gfpNever  <TAB> Never draw the gradient.
     </UL>
     <Related A=PaintGradient>}
  TGFPaintWhen = (gfpAlways, gfpActive, gfpNever);

const
  DEF_PAINT_GRADIENT = gfpAlways;

type
  {: Describes the parameters used by an
     <See Event=TGradientForm.OnCaptionPaint Text=OnCaptionPaint> event handler.
     <BR><BR><B>Sender</B> is the TGradientForm that is being painted.<BR><BR>
     <B>Canvas</B> is the drawing surface that is being painted.  Anything you
     want to appear on the caption must be drawn on this canvas.  This canvas is
     not the actual caption canvas, it is a memory bitmap (non-visible).  This
     prevents flicker as many things are being drawn since the actual visible
     drawing only happens when the entire drawing operation is complete.<BR><BR>
     <B>R</B> is a rectangle that describes the area in which you can draw.
     When the event is first fired, this rectangle will be the entire caption
     less the system icon on the left (if any) and the caption buttons on the
     right (if any).  After performing your drawing operations, this value
     should be modified so that the area you have painted is subtracted out.
     This prevents the gradient from painting over what you have just done.
     <Related A=OnCaptionPaint>}
  TGFOnCaptionPaint = procedure(Sender: TObject; Canvas: TCanvas;
     var R: TRect) of object;

{: TGradientForm is a descendant of the TForm class that paints it's caption
   bar in a gradient fill pattern, like the Microsoft Office applications.
   By default, it starts with black and moves gradually to the system defined
   caption color, although you can override these values.  Also provided is
   an event to allow you to add your own custom painting on the caption bar. }
  TGradientForm = class(TForm)
  private
    // Internal variables
    Colors: array[0..1, 0..MAX_GRADIENT_COLORS-1] of TColorRef;
    CaptionFont: HFONT;
    FGradDefClientProc: TFarProc;
    FGradClientInstance: TFarProc;
    // Property variables
    FCaptionTextColor: TColor;
    FGradientStartColor :TColor;
    FGradientStopColor :TColor;
    FGradientColors: integer;
    FPaintGradient: TGFPaintWhen;
    FCaptionText: string;
    FOnCaptionPaint: TGFOnCaptionPaint;
    FUsingDefaultGradientStopColor: boolean;

    // Internal methods
    function IsActiveWindow: boolean;
    procedure CreateCaptionFont;
    function GetCaptionRect: TRect;
    function DrawCaption(FormDC: HDC; Active: boolean): TRect;
    procedure PaintMenuIcon(DC: HDC; var R: TRect; Active: boolean);
    procedure FillRectSolid(DC: HDC; const R: TRect; Active: boolean);
    procedure FillRectGradient(DC: HDC; const R: TRect; Active: boolean);
    procedure PaintCaptionText(DC: HDC; R: TRect);
    procedure PaintCaptionButtons(DC: HDC; var Rect: TRect);
    // Window message handlers
    procedure WMNCActivate(var Msg: TWMNCActivate); message WM_NCACTIVATE;
    procedure WMNCPaint(var Msg: TMessage); message WM_NCPAINT;
    procedure WMSysColorChange(var Msg: TWMSysColorChange);
       message WM_SYSCOLORCHANGE;
    procedure WMSize(var Msg: TWMSize); message WM_SIZE;
    procedure WMSetCursor(var Msg: TWMSetCursor); message WM_SETCURSOR;
    procedure WMSetText(var Msg: TWMSetText); message WM_SETTEXT;
    procedure WMGetText(var Msg: TWMGetText); message WM_GETTEXT;
    procedure WMGetTextLength(var Msg: TWMGetTextLength);
       message WM_GETTEXTLENGTH;
    procedure WMSettingChange(var Msg: TMessage); message WM_SETTINGCHANGE;
    procedure WMNCLButtonDown(var Msg: TWMNCLButtonDown);
       message WM_NCLBUTTONDOWN;
    procedure WMSysCommand (var Msg: TWMSysCommand); message WM_SYSCOMMAND;
    procedure WMEnterIdle(var Msg: TWMEnterIdle); message WM_ENTERIDLE;
    // MDI Client Window Procedure
    procedure GradClientWndProc(var Message: TMessage);
  protected
    function StoreGradientStopColor: boolean;
    procedure CalculateColors; virtual;
    // Overriden methods
    procedure CreateWnd; override;
    procedure DestroyWnd; override;
    // Property methods
    procedure SetCaptionTextColor(Color: TColor);
    procedure SetGradientStartColor(Color : TColor);
    procedure SetGradientStopColor(Color : TColor);
    procedure SetGradientColors(Val: integer);
    procedure SetPaintGradient(Val: TGFPaintWhen);
    procedure SetCaptionText(const Val: string);
  public
    { This procedure is used to paint the caption gradient. }
    procedure Draw(Active: boolean); virtual;
    // Overridden methods
    { Create creates and initializes an instance of TGradientForm. }
    constructor Create(AOwner: TComponent); override;
    { Destroy destroys an instance of TGradientForm. }
     destructor Destroy; override;
  published
    // Properties
    {: Caption specifies a text string that appears in the caption bar. }
    property Caption: string
       read FCaptionText
       write SetCaptionText
       stored TRUE;
    {: Determines the number of colors used to paint the gradient pattern.  The
       individual colors are determined by fading the start color into the stop
       color.  The number of times this is done is controled by this property.
       The higher the number of colors, the smoother the gradient will appear.
       However, the more colors that are used, the more complex the painting
       will be.
       <Related A=MAX_GRADIENT_COLORS;MIN_GRADIENT_COLORS;DEF_GRADIENT_COLORS> }
    property GradientColors: integer
       read FGradientColors
       write SetGradientColors
       default DEF_GRADIENT_COLORS;
    {: CaptionTextColor is the color that should be used for the text draw in
       the caption bar.  You may have to adjust this color if you change the
       <See Property=TGradientForm.GradientStartColor Text=GradientStartColor>
       to something other than the default of clBlack.
       <Related A=GradientStartColor;GradientStopColor;Caption>}
    property CaptionTextColor: TColor
       read FCaptionTextColor
       write SetCaptionTextColor
       default DEF_CAPTION_TEXT_COLOR;
    {: The leftmost gradient color.  This is the color that is used at the
       beginning of the caption (the far left), and is gradually faded into the
       <See Property=TGradientForm.GradientStopColor Text=GradientStopColor>.
       <Related A=GradientStopColor;GradientColors> }
    property GradientStartColor: TColor
       read FGradientStartColor
       write SetGradientStartColor
       default DEF_GRADIENT_START_COLOR;
    {: The rightmost gradient color. This is the color that is used at the
       end of the caption (the far right), and is gradually faded from the
       <See Property=TGradientForm.GradientStartColor Text=GradientStartColor>.
       <Related A=GradientStartColor;GradientColors> }
    property GradientStopColor: TColor
       read FGradientStopColor
       write SetGradientStopColor
       stored StoreGradientStopColor
       default DEF_GRADIENT_STOP_COLOR;
    {: Determines if and when the gradient caption should be painted.
     <UL>
     <LI>gfpAlways <TAB> The gradient should always be drawn.
     <LI>gfpActive <TAB> Only draw the gradient when the form is active.
     <LI>gfpNever  <TAB> Never draw the gradient.
     </UL> }
    property PaintGradient: TGFPaintWhen
       read FPaintGradient
       write SetPaintGradient
       default DEF_PAINT_GRADIENT;
    {: This event is fired after the icon, buttons and gradient are painted,
       but just before the text is painted.  It is not fired if the caption is
       painted but not as a gradient, that is if
       <See TGradientForm.PaintGradient Text=PaintGradient> is gfpNever or
       gfpActive and the form is not active.<BR><BR> <B>Sender</B> is the
       TGradientForm that is being painted.<BR><BR><B>Canvas</B> is the drawing
       surface that is being painted.  Anything you want to appear on the
       caption must be drawn on this canvas.  This canvas is not the actual
       caption canvas, it is a memory bitmap (non-visible).  This prevents
       flicker as many things are being drawn since the actual visible drawing
       only happens when the entire drawing operation is complete.<BR><BR>
       <B>R</B> is a rectangle that describes the area in which you can draw.
       When the event is first fired, this rectangle will be the entire caption
       less the system icon on the left (if any) and the caption buttons on the
       right (if any).  After performing your drawing operations, this value
       should be modified so that the area you have painted is subtracted out.
       This prevents the gradient from painting over what you have just done. }
    property OnCaptionPaint: TGFOnCaptionPaint
       read FOnCaptionPaint
       write FOnCaptionPaint;
  end;


{$IFDEF DFS_COMPILER_3_UP}
  {: Registers the class for use in the Delphi 3 IDE.  This applies <B>ONLY</B>
     to Delphi 3.  Previous versions of Delphi and C++Builder 1.0 do <B>NOT</>
     support design-time access of TForm descendants.  Sorry, just be happy
     Borland added it to Delphi 3 (and I presume C++B 2.0).  Unlike a normal
     component, TForm descendant classes must also have a package installed for
     the registration process to work.  See the installation notes in
     GradForm.Txt for complete instructions. }
  procedure Register;
{$ENDIF}

implementation

{$R *.DFM}


{$IFDEF DFS_COMPILER_3_UP}
uses
  DsgnIntf;


procedure Register;
begin
  RegisterCustomModule(TGradientForm, TCustomModule);
end;
{$ENDIF}


// A variant record (known as a union in C) to allow easy access to the
// individual red, // green, and blue values of a TColorRef (RGB) value.
type
  TRGBMap = packed record
    case boolean of
      TRUE:  ( RGBVal: DWORD );
      FALSE: ( Red,
               Green,
               Blue,
               Unused: byte );
  end;


var
  EntrancyFlag: boolean;

{: Create creates and initializes an instance of TGradientForm. Call Create
   to instantiate a TGradientForm at runtime.  After calling the inherited
   constructor, Create initializes the following properties:<BR>
   <UL>
   <LI>UsingDefaultGradientStopColor to TRUE.
   <LI>CaptionTextColor to clWhite.
   <LI>GradientStartColor to clBlack.
   <LI>GradientStopColor to clActiveCaption.
   <LI>GradientColors to <See DEF_GRADIENT_COLORS>.
   <LI>PaintGradient to gpfAlways.
   </UL> }
constructor TGradientForm.Create(AOwner: TComponent);
begin
  // We set our new property values first so that they will be valid in the
  // OnCreate event handler.  The inherited Create is what calls that event, so
  // we set up first.
  FUsingDefaultGradientStopColor := TRUE;
  FCaptionTextColor := DEF_CAPTION_TEXT_COLOR;
  // Set gradient start and stop colors.
  FGradientStartColor := DEF_GRADIENT_START_COLOR;
  FGradientStopColor := DEF_GRADIENT_STOP_COLOR;
  // Set the number of colors to use to create the gradient fill.
  FGradientColors := DEF_GRADIENT_COLORS;
  // Should we paint the gradient when window is inactive.
  FPaintGradient := DEF_PAINT_GRADIENT;
  FOnCaptionPaint := NIL;

  // Calculate the colors we need to paint the gradient.
  CalculateColors;
  // Create a font for the caption bar.
  CaptionFont := 0;
  CreateCaptionFont;

  inherited Create(AOwner);
end;


{: Destroy destroys an instance of TGradientForm. Do not call Destroy
   directly in an application. Instead, call Free. Free verifies that the
   instance is not already freed, and only then calls Destroy.<BR>
   Destroy is used to free resources allocated in the
   <See Method=TGradientForm.Create Text=Create> constructor. }
destructor TGradientForm.Destroy;
begin
  // Clean up the font we created.
  if CaptionFont <> 0 then
    DeleteObject(CaptionFont);
  inherited Destroy;
end;

procedure TGradientForm.CreateWnd;
begin
  inherited CreateWnd;
  if FormStyle = fsMDIForm then
  begin
    FGradClientInstance := MakeObjectInstance(GradClientWndProc);
    FGradDefClientProc := Pointer(GetWindowLong(ClientHandle, GWL_WNDPROC));
    SetWindowLong(ClientHandle, GWL_WNDPROC, Longint(FGradClientInstance));
  end;
end;

procedure TGradientForm.DestroyWnd;
begin
  if FormStyle = fsMDIForm then
  begin
    SetWindowLong(ClientHandle, GWL_WNDPROC, Longint(FGradDefClientProc));
    FreeObjectInstance(FGradClientInstance);
  end;
  inherited;
end;

procedure TGradientForm.SetGradientColors(Val: integer);
begin
  if (Val = FGradientColors) or (Val < MIN_GRADIENT_COLORS) or
     (Val > MAX_GRADIENT_COLORS) then
    exit;

  FGradientColors := Val;
  // The number of colors have changed, we need to recalculate the colors we
  // use to paint.
  CalculateColors;
  // Make the non client area repaint.
  SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_DRAWFRAME or SWP_NOACTIVATE or
     SWP_NOMOVE or SWP_NOSIZE or SWP_NOZORDER);
end;

procedure TGradientForm.SetCaptionTextColor(Color: TColor);
begin
  if FCaptionTextColor = Color then
    exit;

  FCaptionTextColor := Color;
  // Make the non client area repaint.
  SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_DRAWFRAME or SWP_NOACTIVATE or
     SWP_NOMOVE or SWP_NOSIZE or SWP_NOZORDER);
end;

procedure TGradientForm.SetGradientStartColor(Color : TColor);
begin
  FGradientStartColor := Color;
  // The colors have changed, we need to recalculate the colors we use to paint.
  CalculateColors;
  // Make the non client area repaint.
  SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_DRAWFRAME or SWP_NOACTIVATE or
     SWP_NOMOVE or SWP_NOSIZE or SWP_NOZORDER);
end;

procedure TGradientForm.SetGradientStopColor(Color : TColor);
begin
  FGradientStopColor := Color;
  FUsingDefaultGradientStopColor := FGradientStopColor = clActiveCaption;
  // The colors have changed, we need to recalculate the colors we use to paint.
  CalculateColors;
  // Make the non client area repaint.
  SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_DRAWFRAME or SWP_NOACTIVATE or
     SWP_NOMOVE or SWP_NOSIZE or SWP_NOZORDER);
end;

procedure TGradientForm.SetPaintGradient(Val: TGFPaintWhen);
begin
  if FPaintGradient = Val then
     exit;

  FPaintGradient := Val;
  // Make the non client area repaint.
  SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_DRAWFRAME or SWP_NOACTIVATE or
     SWP_NOMOVE or SWP_NOSIZE or SWP_NOZORDER);
end;

procedure TGradientForm.SetCaptionText(const Val: string);
begin
  if EntrancyFlag then
    exit;

  EntrancyFlag := TRUE;
  try
    inherited Caption := Val; { Need this for WindowMenu items in MDI }
  finally
    EntrancyFlag := FALSE;
  end;

  if Val = FCaptionText then
    exit;

  FCaptionText := Val;
  if ((FormStyle = fsMDIChild) and (WindowState = wsMaximized)) then
    // Need to cause main form's caption to be redrawn, not the MDI child.
    SetWindowPos(Application.MainForm.Handle, 0, 0, 0, 0, 0, SWP_DRAWFRAME or
       SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOSIZE or SWP_NOZORDER)
  else
    Draw(IsActiveWindow);
end;

function TGradientForm.IsActiveWindow: boolean;
begin
  if FormStyle = fsMDIChild then
    if assigned(Application.MainForm) then
      Result := (GetActiveWindow = Application.MainForm.Handle) and
                (TForm(Application.MainForm).ActiveMDIChild = Self)
    else
      Result := FALSE
  else
    Result := GetActiveWindow=Handle;
end;

procedure TGradientForm.CalculateColors;
var
  LoColor, HiColor: TRGBMap;
  RedPct,
  GreenPct,
  BluePct: real;
  x,
  Band: integer;
begin
  // Get colors for both active and inactive captions.
  for x := 0 to 1 do
  begin
    if x = 0 then   // inactive captions
    begin
      LoColor.RGBVal := RGB(0, 0, 0);;
      HiColor.RGBVal := GetSysColor(COLOR_INACTIVECAPTION);
    end else begin    // active caption
      LoColor.RGBVal := ColorToRGB(FGradientStartColor);
      HiColor.RGBVal := ColorToRGB(FGradientStopColor);
    end;
    // Figure out the percentage of each RGB value needed for banding
    RedPct   := (HiColor.Red - LoColor.Red)/ (FGradientColors-1);
    GreenPct := (HiColor.Green - LoColor.Green) / (FGradientColors-1);
    BluePct  := (HiColor.Blue - LoColor.Blue) / (FGradientColors-1);
    // Use the percentage of each color to create each band color.
    for Band := 0 to (FGradientColors-1) do
      Colors[x][Band] := RGB(LoColor.Red + round(RedPct * (Band)),
         LoColor.Green + round(GreenPct * (Band)),
         LoColor.Blue + round(BluePct * (Band)));
  end;
end;

procedure TGradientForm.CreateCaptionFont;
var
  NCM: TNonClientMetrics;
begin
  if CaptionFont <> 0 then
    DeleteObject(CaptionFont);
  NCM.cbSize := SizeOf(NCM);
  if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, @NCM, 0) then
  begin
    if BorderStyle in [bsToolWindow, bsSizeToolWin] then
      CaptionFont := CreateFontIndirect(NCM.lfSmCaptionFont)
    else
      CaptionFont := CreateFontIndirect(NCM.lfCaptionFont);
  end else
    CaptionFont := 0;
end;

// The caption rect is the rectangle we are interested in painting.  This will
// be the area that contains the caption icon, text and buttons.
function TGradientForm.GetCaptionRect: TRect;
begin
  // Designing mode always draws the form as bsSizeable
  if csDesigning in ComponentState then
  begin
    GetWindowRect(Handle, Result);
    // Convert rect from screen (absolute) to client (0 based) coordinates.
    OffsetRect(Result, -Result.Left, -Result.Top);
    // Shrink rectangle to allow for window border.  We let Windows paint it.
    InflateRect(Result, -GetSystemMetrics(SM_CXSIZEFRAME),
       -GetSystemMetrics(SM_CYSIZEFRAME));
    Result.Bottom := Result.Top + GetSystemMetrics(SM_CYCAPTION) - 1;
  end else begin
    // if we have no border style, then just set the rectange empty.
    if BorderStyle = bsNone then
      SetRectEmpty(Result)
    else begin
      GetWindowRect(Handle, Result);
      // Convert rect from screen (absolute) to client (0 based) coordinates.
      OffsetRect(Result, -Result.Left, -Result.Top);
      // Shrink rectangle to allow for window border.  We let Windows paint it.
      case BorderStyle of
        bsToolWindow, bsSingle, bsDialog:
          InflateRect(Result, -GetSystemMetrics(SM_CXFIXEDFRAME),
             -GetSystemMetrics(SM_CYFIXEDFRAME));
        bsSizeable, bsSizeToolWin:
          InflateRect(Result, -GetSystemMetrics(SM_CXSIZEFRAME),
             -GetSystemMetrics(SM_CYSIZEFRAME));
      end;

      // Set the appropriate height of caption bar.
      if BorderStyle in [bsToolWindow, bsSizeToolWin] then
        Result.Bottom := Result.Top + GetSystemMetrics(SM_CYSMCAPTION) - 1
      else
        Result.Bottom := Result.Top + GetSystemMetrics(SM_CYCAPTION) - 1;
    end;
  end;
end;

// Paint the icon for the system menu.
procedure TGradientForm.PaintMenuIcon(DC: HDC; var R: TRect; Active: boolean);
{$IFDEF DFS_COMPILER_2}
const
  LR_COPYFROMRESOURCE = $4000; // Missing from WINDOWS.PAS in Delphi 2!
{$ENDIF}
var
  SmallCopy,
  IconHandle: HIcon;
  Size: integer;
begin
  // Does the form have an icon assigned to it?
  if Icon.Handle <> 0 then
    IconHandle := Icon.Handle
  // If not, does the application have an icon?
  else if Application.Icon.Handle <> 0 then
    IconHandle := Application.Icon.Handle
  // If not, then just use the system defined application icon.
  else
    IconHandle := LoadIcon(0, IDI_APPLICATION);

  Size := GetSystemMetrics(SM_CXSMICON);
  SmallCopy := CopyImage(IconHandle, IMAGE_ICON, Size, Size,
     LR_COPYFROMRESOURCE);
  with R do
    // Let CopyImage() make get us a nice 16x16 version of the icon and we'll
    // paint it.
    DrawIconEx(DC, Left+1, Top+1, SmallCopy, 0, 0, 0, 0, DI_NORMAL);

  DestroyIcon(SmallCopy);
  Inc(R.Left, Size+1);
end;

// Paint the given rectangle with the system solid color.
procedure TGradientForm.FillRectSolid(DC: HDC; const R: TRect; Active: boolean);
var
  OldBrush,
  Brush: HBrush;
begin
  // Create a brush with the appropriate color\
  if Active then
    Brush := CreateSolidBrush(GetSysColor(COLOR_ACTIVECAPTION))
  else
    Brush := CreateSolidBrush(GetSysColor(COLOR_INACTIVECAPTION));
  // Select that brush into the temporary DC.
  OldBrush := SelectObject(DC, Brush);
  try
    // Fill the rectangle using the selected brush -- PatBlt is faster than
    // FillRect
    with R do
      PatBlt(DC, Left, Top, Right-Left, Bottom-Top, PATCOPY);
  finally
    // Clean up the brush
    SelectObject(DC, OldBrush);
    DeleteObject(Brush);
  end;
end;

// Paint the given rectangle with the gradient pattern.
procedure TGradientForm.FillRectGradient(DC: HDC; const R: TRect;
   Active: boolean);
var
  OldBmp,
  TmpBmp: HBitmap;
  TmpDC: HDC;
  OldBrush,
  Brush: HBrush;
  Step: real;
  Band: integer;
  H: integer;
begin
  // This may look backwards, but it's not.  If the device capabilities
  // indicate that there are palette entries (more than 0), then we are in
  // a low color mode.  This is because when in high color mode or above,
  // Windows doesn't use palettes; each pixel knows it's RGB value.
  if (GetDeviceCaps(DC, SIZEPALETTE) > 0) or
     ((R.Right - R.Left) < GradientColors) then
  begin
    // Low color gradient, slower

    // Determine how large each band should be in order to cover the
    // rectangle (one band for every color intensity level).
    Step := (R.Right - R.Left) / FGradientColors;
    H := R.Bottom - R.Top;

    // Start filling bands
    for Band := 0 to (FGradientColors-1) do
    begin
      // Create a brush with the appropriate color for this band
      Brush := CreateSolidBrush(Colors[ord(Active)][Band]);
      // Select that brush into the temporary DC.
      OldBrush := SelectObject(DC, Brush);
      try
        // Fill the rectangle using the selected brush -- PatBlt is faster than
        // FillRect
        PatBlt(DC, round(Band*Step), 0, round((Band+1)*Step), H, PATCOPY);
      finally
        // Clean up the brush
        SelectObject(DC, OldBrush);
        DeleteObject(Brush);
      end;
    end; // for
  end else begin
    // High color gradient, faster
    H := R.Bottom - R.Top;

    TmpDC := CreateCompatibleDC(DC);
    TmpBmp := CreateCompatibleBitmap(DC, FGradientColors, 1);
    OldBmp := SelectObject(TmpDC, TmpBmp);
    try
      // Start filling bands
      for Band := 0 to (FGradientColors-1) do
        SetPixel(TmpDC, Band, 0, Colors[ord(Active)][Band]);
      StretchBlt(DC, 0, 0, R.Right-R.Left, H, TmpDC, 0, 0, FGradientColors-1, 1,
         SRCCOPY);
    finally
      SelectObject(TmpDC, OldBmp);
      DeleteObject(TmpBmp);
      DeleteDC(TmpDC);
    end;
  end;
end;

procedure TGradientForm.PaintCaptionText(DC: HDC; R: TRect);
var
  OldColor: TColorRef;
  OldMode: integer;
  OldFont: HFont;
  CaptionText: string;
begin
  CaptionText := Caption;
  // Have to turn off complete boolean eval for this "if" statement.  I never
  // have it on anyway, but some do.
  {$IFOPT B+} {$DEFINE DFS_RESET_BOOL_EVAL} {$B-} {$ENDIF}
  if ((FormStyle = fsMDIForm) and (ActiveMDIChild <> NIL) and
      (ActiveMDIChild.WindowState = wsMaximized)) then
    CaptionText := CaptionText + ' - [' + ActiveMDIChild.Caption + ']';
  {$IFDEF DFS_RESET_BOOL_EVAL} {$B+} {$UNDEF DFS_RESET_BOOL_EVAL} {$ENDIF}

  Inc(R.Left, 2);
  // Set the text color to white, not the system color.  We always color from
  // black to the system color, so we will be almost guaranteed of having a
  // nearly black background under the text.  White is the safest color to use
  // in this case (Microsoft thinks so, too, so I can't be too wrong).
  OldColor := SetTextColor(DC, ColorToRGB(FCaptionTextColor));
  // Set the background text painting mode to transparent so that drawing text
  // doesn't distrub the gradient we just painted.  If you didn't do this, then
  // drawing text would also fill the text rectangle with a solid background
  // color, screwing up our gradient.
  OldMode := SetBkMode(DC, TRANSPARENT);
  // Select in the system defined caption font (see Create constructor).
  if CaptionFont <> 0 then
    OldFont := SelectObject(DC, CaptionFont)
  else
    OldFont := 0;
  try
    // Draw the text making it left aligned, centered vertically, allowing no
    // line breaks.
    DrawText(DC, PChar(CaptionText), -1, R, DT_LEFT or DT_VCENTER or
       DT_SINGLELINE or DT_END_ELLIPSIS);
  finally
    // Clean up all the drawing objects.
    if OldFont <> 0 then
      SelectObject(DC, OldFont);
    SetBkMode(DC, OldMode);
    SetTextColor(DC, OldColor);
  end;
end;

// Paint the min/max/help/close buttons.
procedure TGradientForm.PaintCaptionButtons(DC: HDC; var Rect: TRect);
var
  BtnWidth: integer;
  Flag: UINT;
  SrcRect: TRect;
  ABorderStyle: TFormBorderStyle;
  ABorderIcons: TBorderIcons;
begin
  SrcRect := Rect;
  InflateRect(SrcRect, -2, -2);
  if csDesigning in ComponentState then
  begin
    // While designing, the min/max buttons are always shown in a sizeable frame
    ABorderStyle := bsSizeable;
    ABorderIcons := [biSystemMenu, biMinimize, biMaximize];
  end else begin
    ABorderStyle := BorderStyle;
    ABorderIcons := BorderIcons;
  end;

  if ABorderStyle in [bsToolWindow, bsSizeToolWin] then
  begin
    // Tool windows only have the close button, nothing else.
    with SrcRect do
      Left := Right - (GetSystemMetrics(SM_CXSMSIZE)) + 2;
    DrawFrameControl(DC, SrcRect, DFC_CAPTION, DFCS_CAPTIONCLOSE);
    Rect.Right := SrcRect.Left-5;
  end else begin
    BtnWidth := GetSystemMetrics(SM_CXSMICON)-2;
    { Windows is loopy.  It always returns an even number, no matter what }
    if (Odd(BtnWidth) XOR Odd(Rect.Bottom-Rect.Top)) then
      inc(BtnWidth);
    SrcRect.Left := SrcRect.Right - BtnWidth - 2;
    // if it has system menu, it has a close button.
    if biSystemMenu in ABorderIcons then
    begin
      DrawFrameControl(DC, SrcRect, DFC_CAPTION, DFCS_CAPTIONCLOSE);
      OffsetRect(SrcRect, -BtnWidth-4, 0);
      Dec(Rect.Right,BtnWidth+4);
    end;
    // Minimize and Maximized don't show up at all if BorderStyle is bsDialog or
    // if neither of them are enabled.
    if (ABorderStyle in [bsSizeable, bsSingle]) and
       (ABorderIcons * [biMinimize, biMaximize] <> []) then
    begin
      if WindowState = wsMaximized then
        Flag := DFCS_CAPTIONRESTORE
      else
        Flag := DFCS_CAPTIONMAX;
      // if it doesn't have max in style, then it shows up disabled.
      if not (biMaximize in ABorderIcons) then
        Flag := Flag or DFCS_INACTIVE;

      DrawFrameControl(DC, SrcRect, DFC_CAPTION, Flag);
      OffsetRect(SrcRect, -BtnWidth-2, 0);
      Dec(Rect.Right,BtnWidth+2);

      Flag := DFCS_CAPTIONMIN;
      // if it doesn't have min in style, then it shows up disabled.
      if not (biMinimize in ABorderIcons) then
        Flag := Flag or DFCS_INACTIVE;

      DrawFrameControl(DC, SrcRect, DFC_CAPTION, Flag);
      OffsetRect(SrcRect, -BtnWidth-2, 0);
      Dec(Rect.Right,BtnWidth+2);
    end;

    // Help only shows up in bsDialog style, and bsSizeable, bsSingle when there
    // is no min or max button.
    if biHelp in ABorderIcons then
    begin
      if ((ABorderStyle in [bsSizeable, bsSingle]) and
         (ABorderIcons * [biMinimize, biMaximize] = [])) or
         (ABorderStyle = bsDialog) then
      begin
        DrawFrameControl(DC, SrcRect, DFC_CAPTION, DFCS_CAPTIONHELP);
        Dec(Rect.Right,BtnWidth+2);
      end;
    end;

    Dec(Rect.Right, 3);
  end;
end;


function TGradientForm.DrawCaption(FormDC: HDC; Active: boolean): TRect;
var
  R: TRect;
  OldBmp,
  Bmp: HBitmap;
  BmpDC: HDC;
  BmpCanvas: TCanvas;
  w,h:integer;
begin
  // Get only the portion we need to draw.
  R := GetCaptionRect;
  Result := R;

  // Convert to logical (0-based) coordinates
  OffsetRect(R, -R.Left, -R.Top);

  W := R.Right - R.Left;
  H := R.Bottom - R.Top;

  // Create a temporary device context to draw on.  Drawing on a temporary DC
  // and copying it to the real DC accomplishes two things:
  // 1) It is faster because Windows doesn't have to draw anything in the
  //    temporary DC on the screen, it only draws when you paint something on a
  //    real DC.  Then it just draws everything at once when we copy it, instead
  //    of drawing a little, do some calculations, draw a little, etc.
  // 2) It looks much better because it is drawn faster.  It reduces the
  //    "flicker" that you would see from each individual part being drawn,
  //    especially the gradient bands.
  BmpDC := CreateCompatibleDC(FormDC);
  Bmp := CreateCompatibleBitmap(FormDC, W, H);
  OldBmp := SelectObject(BmpDC, Bmp);

  try
    if (FPaintGradient = gfpAlways) or
       (Active and (FPaintGradient = gfpActive)) then
      // Draw the gradient background in the temporary DC
      FillRectGradient(BmpDC, R, Active)
    else
      FillRectSolid(BmpDC, R, Active);

    Inc(R.Left, 1);
    // Do we need to paint an icon for the system menu?
    if not ((FormStyle = fsMDIChild) and (WindowState = wsMaximized)) then
    begin
      if (biSystemMenu in BorderIcons) and
         (BorderStyle in [bsSingle, bsSizeable]) then
        // PaintMenuIcon will adjust the rect so that future drawing operations
        // happen in the right spot.
        PaintMenuIcon(BmpDC, R, Active);

      PaintCaptionButtons(BmpDC, R); // Paint the min/max/help/close buttons.
    end;
    if assigned(FOnCaptionPaint) then
    begin
      BmpCanvas := TCanvas.Create;
      try
        BmpCanvas.Handle := BmpDC;
        BmpCanvas.Font.Handle := CaptionFont;
        FOnCaptionPaint(Self, BmpCanvas, R);
      finally
        BmpCanvas.Free;
      end;
    end;
    PaintCaptionText(BmpDC, R); // Paint the caption text.

    // Copy the gradient caption bar to the real DC.
    BitBlt(FormDC, Result.Left, Result.Top, W, H, BmpDC, 0, 0, SRCCOPY);
  finally
    // Clean up all the temporary drawing objects.
    SelectObject(BmpDC, OldBmp);
    DeleteObject(Bmp);
    DeleteDC(BmpDC);
  end;
end;

// Windows sends this message when the window has been activated or deactivated.
procedure TGradientForm.WMNCActivate(var Msg: TWMNCActivate);
begin
  Msg.Result := 1;
  inherited; { Call inherited or bad things will happen with MDI }
  Draw(Msg.Active);

{ Don't need this stuff any more, but I'm saving it just in case... }

(* we always need to paint it now since we are managing the caption text.
  // Do we need to paint a gradient?
  if Msg.Active or FGradientOnInactive then begin
    Draw(Msg.Active);
    // Tell Windows we have handled the message and it doesn't need to do
    // anything else.
    Msg.Result := 1;
  end else
    // Let Windows paint the caption.
    inherited;
*)
end;

// Windows sends this message whenever any part of the non-client area
// (caption, window border) needs repainting.
procedure TGradientForm.WMNCPaint(var Msg: TMessage);
var
  WR, R: TRect;
  DC: HDC;
  MyRgn: HRGN;
  DeleteRgn: boolean;
begin
  DeleteRgn := FALSE;
  // The region that needs painting is passed in WParam.  A region is a Windows
  // object used to describe the non-rectangular area used by a combination of
  // rectangles.
  MyRgn := Msg.wParam;
  DC := GetWindowDC(Handle);
  try
    GetWindowRect(Handle, WR);
    // Select the update region as the clipping region.  Clipping regions
    // guarantee that any painting done outside of the selected region is not
    // shown (thrown away).
    if SelectClipRgn(DC, MyRgn) = ERROR then
    begin
      // We got passed an invalid region.  Generally, this happens when the
      // window is first created or a MDI is minimized.  We'll create our own
      // region (the rectangle that makes up the entire window) and use that
      // instead.
      with WR do
        MyRgn := CreateRectRgn(Left, Top, Right, Bottom);
      SelectClipRgn(DC, MyRgn);
      DeleteRgn := TRUE;
    end;
    // Convert the clipping region coordinates from screen to client.
    OffsetClipRgn(DC, -WR.Left, -WR.Top);
    // Draw our gradient caption.
    R := DrawCaption(DC, IsActiveWindow);
    // Here's the trick.  DrawCaption returns the rectangle that we painted.
    // We now exclude that rectangle from the clipping region.  This guarantees
    // that any further painting that occurs will not happen in this rectangle.
    // That means that when we let the default painting for WM_NCPAINT occur,
    // it will not paint over our gradient. It only paints the stuff that we
    // didn't, like the window borders.
    ExcludeClipRect(DC, R.Left, R.Top, R.Right, R.Bottom);
    // Convert coordinates back into screen-based.
    OffsetClipRgn(DC, WR.Left, WR.Top);
    // Get the region that is now described by the clipping region.
    GetClipRgn(DC, MyRgn);
    // Pass that region on to the default WM_NCPAINT handler.  Remember, we
    // excluded the rectangle that we painted, so Windows will not be able to
    // paint over what we did. Most gradient captions components just let
    // windows draw its stuff first, and then paint the gradient.  This results
    // in an irritating "flicker", caused by the area being painted normally,
    // and then painted over a second time by the gradient.
    Msg.Result := DefWindowProc(Handle, Msg.Msg, MyRgn, Msg.lParam);
  finally
    // If we had to create our own region, we have to clean it up.
    if DeleteRgn then
      DeleteObject(MyRgn);
    ReleaseDC(Handle, DC); // NEVER leave this hanging.
  end;
end;

// Windows sends this message if the user changes any of the system colors.
procedure TGradientForm.WMSysColorChange(var Msg: TWMSysColorChange);
begin
  if FUsingDefaultGradientStopColor then
    FGradientStopColor := clActiveCaption;
  CalculateColors;
  inherited;
end;

// The window has been resized.
procedure TGradientForm.WMSize(var Msg: TWMSize);
begin
  inherited;
  // If the window was maximized or restored, we need to redraw so the right
  // caption button is painted.
  if (Msg.SizeType = SIZE_MAXIMIZED) or (Msg.SizeType = SIZE_RESTORED) then
    Draw(IsActiveWindow);
end;

// Windows would like to have a cursor displayed.  I know, you're wondering
// why the hell I care about this, aren't you?  Well, in the inherited handling
//  (default Windows processing) of this message, if the mouse is over a
// resizeable border section, Windows repaints the caption buttons.  Why?  I
// have absolutely no idea.  However, that's not the important part.  When it
// repaints those buttons, it also repaints the background around them in the
// last color it painted the caption in.  Now, usually this would just result
// in losing a few bands of the caption gradient, which 99.44% of all users
// would never notice.  However, because we don't always allow default
// processing of WM_NCACTIVATE, sometimes Windows doesn't have the right idea
// about which color is currently the background.  This cause the background to
// get painted in the wrong color sometimes, which 99.44% of all users *will*
// notice.  We fix it by setting the appropriate cursor and not allowing the
// default processing to occur.
procedure TGradientForm.WMSetCursor(var Msg: TWMSetCursor);
begin
  // Tell Windows we handled the message
  Msg.Result := 1;
  // Load and display the correct cursor for the border area being hit
  case Msg.HitTest of
    HTTOP,
    HTBOTTOM:      SetCursor(LoadCursor(0, MakeIntResource(IDC_SIZENS)));
    HTLEFT,
    HTRIGHT:       SetCursor(LoadCursor(0, MakeIntResource(IDC_SIZEWE)));
    HTTOPRIGHT,
    HTBOTTOMLEFT:  SetCursor(LoadCursor(0, MakeIntResource(IDC_SIZENESW)));
    HTTOPLEFT,
    HTBOTTOMRIGHT: SetCursor(LoadCursor(0, MakeIntResource(IDC_SIZENWSE)));
  else
    // Wasn't anything we cared about, so tell Windows we didn't handle it.
    Msg.Result := 0;
    inherited;
  end;
end;


procedure TGradientForm.WMSetText(var Msg: TWMSetText);
const
  FirstTime: boolean = TRUE;
begin
  // The dilema:  If we call inherited, it will cause horrible flicker in the
  // caption bar.  If we don't call inherited, the underly HWND never gets it's
  // window text value changed, and thus things like Alt+Tabbing show the old
  // value.  I've found through trial and error, wild guessing, pure luck and
  // plain old pull-it-outta-your-butt magic that I can stop the flickering by
  // clearing the redraw flag before calling inherited and then resetting after.
  // But don't do it if it isn't visible or its horrible when first opening the
  // form.
  if Visible then
    SendMessage(Handle, WM_SETREDRAW, 0, 0);
  try
    inherited;
  finally
    if Visible then
      SendMessage(Handle, WM_SETREDRAW, 1, 0);
  end;
  Caption := Msg.Text;
end;

procedure TGradientForm.WMGetText(var Msg: TWMGetText);
begin
  StrLCopy(Msg.Text, PChar(FCaptionText), Msg.TextMax-1);
  Msg.Result := StrLen(Msg.Text)+1;
end;

procedure TGradientForm.WMGetTextLength(var Msg: TWMGetTextLength);
begin
  Msg.Result := Length(FCaptionText);
end;

procedure TGradientForm.WMSettingChange(var Msg: TMessage);
begin
  // User might have changed NC font.
  if Msg.wParam = SPI_SETNONCLIENTMETRICS then
    CreateCaptionFont;
  inherited;
end;

{: This procedure is used to paint the caption gradient.  It is normally
   called internally, but it can be used any time a repaint of the caption
   is needed. The <B>Active</B> parameter is used to indicate whether the
   caption should be painted as the active window or an inactive window. }
procedure TGradientForm.Draw(Active: boolean);
var
  DC: HDC;
begin
  // Get the DC we need to paint in.  GetDC would only get the DC for the
  // client area, we need it for non-client area, too, so we use GetWindowDC.
  DC := GetWindowDC(Handle);
  try
    DrawCaption(DC, Active);
  finally
    ReleaseDC(Handle, DC); // NEVER leave this hanging.
  end;
end;

procedure TGradientForm.GradClientWndProc(var Message: TMessage);
begin
  with Message do begin
    Result := CallWindowProc(FGradDefClientProc, ClientHandle, Msg, wParam,
       lParam);
    // if you don't want your MDI child forms to be TGradientForm descendants,
    // you will need to use uncomment the code in the following line.
    // It will work, but it causes an annoying flicker.
    // NOTE:  as of v1.55, this is no longer necessary.  Just leave it the way
    //        it is....I think.....

    // $003F is a "Magic Number".  I hate this, but it's the only way I could
    // get it to work.  :(
    if {(Msg = WM_MDIREFRESHMENU) or} (Msg = $003F) then
      Draw(IsActiveWindow);
  end;
end;

procedure TGradientForm.WMNCLButtonDown(var Msg: TWMNCLButtonDown);
begin
  inherited;
  Draw(IsActiveWindow);
end;

procedure TGradientForm.WMSysCommand (var Msg : TWMSysCommand);
begin
  Draw(IsActiveWindow);
  inherited;
  Draw(IsActiveWindow);
end;

procedure TGradientForm.WMEnterIdle(var Msg: TWMEnterIdle);
begin
  Draw(IsActiveWindow);
  inherited;
end;

function TGradientForm.StoreGradientStopColor: boolean;
begin
  Result := not FUsingDefaultGradientStopColor;
end;

initialization
  EntrancyFlag := FALSE;
end.

